/* 
 * Script umka220.p
 * ver 1.1.0
 * (c) Copyright 2022, GLONASSsoft
 * This file is provided as is (no warranties).
 *
 * Скрипт работы со считывателем УМКа220 по интерфейсу RS485
 *
 * Описание: скрипт осуществляет считывание счетчика УМКа220 по интерфейсу RS485 с сохранением значения в EEPROM.
 * Значение счетчика через параметры скрипт выдает беззнаково, подобно счетчикам типа УСС. 
 *
 * ВНИМАНИЕ! Настоятельно НЕ рекомендуется применять терминалы УМКа3ХХ в комплектациях без внутреннего аккумулятора
 * для работы с данным скриптом, так как в случае перезагрузки терминала в его EEPROM могут быть сохранены
 * неактуальные данные со счетчика.
 *
 * Настройка скорости обмена по шине RS485 осуществляется на вкладке конфигуратора "Интерфейсы"
 *
 * Поддерживаемые команды:
 * chat GETADDR - получить значение адреса счетчика, с которым производится работа
 * chat GETGCOUNT - получить накопленное в EEPROM значение счетчика в знаковом виде
 * chat GETCOUNT - получить значение счетчика с его последней перезагрузки в знаковом виде
 * chat SETADDR=n - установить значение адреса счетчика (n - адрес, значения: от 1 до 254 включительно)
 *
 */

#include <time>
#include <string>
#include tracker
#include serial
#include crc

#define TICK            1000
#define MAGIC           0x55323230          // "U220"
#define PORT            SERIAL_PORT_0
#define COUNTER_TAG     0

// дефолтные значения настроек
#define DEF_ADDR        0x01

// адреса параметров конфигурации в eeprom
#define MAGIC_CELL      0x00
#define ADDR_CELL       0x01
#define PCOUNT_CELL     0x02
#define GCOUNT_CELL     0x03
#define DT_CELL         0x04

// команды
#define R_COUNT_COM     0x0500
#define R_UPTIME_COM    0x0700

#pragma dynamic 128

//---------------------------------------------------
new config[
  .addr,
  .pcount,
  .gcount,
  .deltat
]

new counter = 0
new uptime = 0
new bool:count_valid = false

//---------------------------------------------------
get_config()
{
  config.addr   = DEF_ADDR
  config.pcount = 0
  config.gcount = 0
  config.deltat = 0
  
  if (geteep(MAGIC_CELL) == MAGIC)
  {
    config.addr   = geteep(ADDR_CELL)
    config.pcount = geteep(PCOUNT_CELL)
    config.gcount = geteep(GCOUNT_CELL)
    config.deltat = geteep(DT_CELL)
  }
}

//---------------------------------------------------
save_config()
{
  seteep(ADDR_CELL, config.addr)
  seteep(PCOUNT_CELL, config.pcount)
  seteep(GCOUNT_CELL, config.gcount)
  seteep(DT_CELL, config.deltat)
  seteep(MAGIC_CELL, MAGIC)
}

//----- отправить команду и получить ответ ----------
bool: send_command(addr, com, &value)
{
  new buff[16] = 0x00
  new len
  new crc

  // отправка запроса
  buff[0] = addr            // адресс
  buff[1] = 0x03            // длинна запроса
  buff[2] = 0xDC            // префикс
  buff[3] = com >> 8        // код комманды
  buff[4] = com & 0xFF      // код подкомманды
  crc = calc_crcX(buff,5)
  buff[5] = crc
  rssend(buff, 6)
  
  // прием ответа
  len = rsrecv(buff, 15, 300, 10)
  if (len == 11 && check_crcX(buff, len) && buff[2] == 0xDC && buff[0] == addr)
  {
    value = (buff[9] & 0xFF) + ((buff[8] & 0xFF) << 8) + ((buff[7] & 0xFF) << 16) + ((buff[6] & 0xFF) << 24)
    return true
  }
  return false
}

//===================================================
main()
{
  get_config()
  settimer(TICK)
}

//===================================================
@timer()
{
  new unixtime = 0
  
  count_valid = false
  unixtime = gettime()
  
  if (unixtime)                 // текущее unixtime успешно получено (!= 0)
  {
    if (rsopen(SB_AUTO, PORT))
    {
      count_valid = send_command(config.addr, R_UPTIME_COM, uptime)
      count_valid = send_command(config.addr, R_COUNT_COM, counter) && count_valid
      rsclose()
      if (count_valid)          // получены новые данные
      {
        new dt = unixtime - uptime                             // получаем текущую дельту времен
        printf("%d, %d", config.deltat, dt)
        if (dt > config.deltat + 1) config.gcount += counter   // если текущая дельта явно больше нынешней - была перезагрузка 220 и предыдущее значение счетчика 220 не актуально
        else config.gcount += counter - config.pcount          // иначе добавляем только разницу текущего полученного с предыдущим
        config.deltat = dt                                     // сохраняем текущую дельту
        config.pcount = counter                                // сохраняем полученный счетчик
        save_config()                                          // все что нужно - в eeprom
      }
    }
  }
  settag(COUNTER_TAG, config.gcount, count_valid)
  printf("%d", config.gcount)
}

//===================================================
@chat(string{128})
{
  // запрос значения адреса
  if (strequal(string, "GETADDR", true, 7))             // пример комманды > chat GETADDR
  {
    strformat(string, sizeof(string), true, "ADDR=%d", config.addr)
  }
  // запрос значения счетчика за все время
  else if (strequal(string, "GETGCOUNT", true, 9))      // пример комманды > chat GETGCOUNT
  {
    strformat(string, sizeof(string), true, "GCOUNT=%d", config.gcount)
  }
  // запрос значения счетчика в текущей сессии
  else if (strequal(string, "GETCOUNT", true, 8))       // пример комманды > chat GETCOUNT
  {
    if (count_valid) strformat(string, sizeof(string), true, "COUNT=%d", counter)
    else strcopy(string, "COUNT not valid")
  }
  // установка адреса
  else if (strequal(string, "SETADDR", true, 7))        // пример комманды > chat SETADDR=5
  {
    new NewValue = 0
    if (strlen(string) >= 9)
    {
      NewValue = strval(string, 8)
      if (1 <= NewValue <= 254)
      {
        config.addr = NewValue
        strformat(string, sizeof(string), true, "ADDR=%d", NewValue)
        save_config()
      }
      else strcopy(string, "ADDR out of range [1, 254]")
    }
    else strcopy(string, "parameter missed '=?'")
  }
  // получена неизвестная команда
  else
    strcopy(string, "Unknown command")
}